from .xgb import XGB
from .cat import Cat
from .elastic_net import ElasticNetM
from .gradient_boosting import GradientBoosting
from .lasso import LassoM
from .lgb import LGB
from .sgd import SGD
from .svr import SVRM


from sklearn.naive_bayes import MultinomialNB
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.model_selection import train_test_split, GridSearchCV
from ..data_manipulation.metrics import Metrics
import numpy as np

class Baseline():
  def __init__(self) -> None:
    self.model = None
    self.parameters = None

  def create_raw(self, X_raw, y_raw,params=None, cv=5):
    pipeline = Pipeline([
        ('vect', CountVectorizer()),
        ('clf', MultinomialNB())
    ])
    params_columns = ['vect__max_features','vect__strip_accents','vect__min_df',
                      'vect__ngram_range','clf__alpha']
    
    params_basic = {
        'vect__max_features': [100, 1000],
        'vect__strip_accents': ['ascii'],
        'vect__min_df': np.linspace(0.0001,0.3, 3),
        'vect__ngram_range': [(1, 1), (1, 2),(1,3),(2,1)],
        'clf__alpha': np.linspace(0.001,1,10)
    }
    if params == None:
            params = params_basic
    else:
        for parameter in params_columns:
            if parameter not in params.keys():
                params[parameter] = params_basic[parameter]

    grid_search = GridSearchCV(pipeline, params, cv=cv, n_jobs=-1)
    grid_search.fit(X_raw, y_raw)
    self.parameters = grid_search.best_params_
    self.model = grid_search.best_estimator_


  def create(self, X, y, params, cv=5):
    params_columns = ['alpha']
    
    params_basic = {
        'alpha': np.linspace(0.001,1,10)
    }
    if params == None:
            params = params_basic
    else:
        for parameter in params_columns:
            if parameter not in params.keys():
                params[parameter] = params_basic[parameter]

    bayes = MultinomialNB()
    grid_search = GridSearchCV(bayes, params, cv=cv, n_jobs=-1)
    grid_search.fit(X, y)
    self.parameters = grid_search.best_params_
    self.model = grid_search.best_estimator_

  def get(self):
    return self.model

    




from sklearn.metrics import mean_squared_error,r2_score
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.model_selection import GridSearchCV, train_test_split,KFold
import optuna.visualization as vis
from ..metrics_regression import Metrics
from joblib import dump, load
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class GradientBoosting(Metrics):
    def __init__(self):
        self.model = None
        self.parameters = None

    def put(self, model):
        self.model = model

    def create(self, X, y, params=None):
        if params == None:
          gbm = GradientBoostingRegressor()
          gbm.fit(X,y)
          self.model = gbm
        else:
            gbm = GradientBoostingRegressor(**params)
            gbm.fit(X,y)
            self.model = gbm
            self.parameters = params
          
    def create_grid(self, X, y,params=None, scoring='neg_mean_squared_error', n_jobs =-1, cv=5):
        params_columns = ['learning_rate', 'n_estimators', 'max_depth',
                        'min_samples_split', 'min_samples_leaf', 'max_features',
                        'subsample', 'loss', 'alpha', 'random_state', 'validation_fraction']
        params_basic = {
            'learning_rate': [0.01, 0.1],
            'n_estimators': [10, 200],
            'max_depth': [1,10],
            'min_samples_split': [10,30],
            'min_samples_leaf': [10,100],
            'max_features': [0.5, 'sqrt', 10],
            'subsample': [0.5, 1],
            'loss': ['ls', 'lad', 'huber', 'quantile'],
            'alpha': [0.1, 0.9],
            'random_state': [42],
            'validation_fraction': [0.2]
            }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        gbm = GradientBoostingRegressor()
        grid_search = GridSearchCV(gbm, params, cv=cv, scoring=scoring,n_jobs=n_jobs)
        grid_search.fit(X, y)
        best_model = grid_search.best_estimator_
        self.model = best_model
        self.parameters = grid_search.best_params_

    def create_optuna(self, X, y, params=None, n_trials=5, show_plot=False, show_features=False):
        params_columns = ['learning_rate', 'n_estimators', 'max_depth',
                        'min_samples_split', 'min_samples_leaf', 'max_features',
                        'subsample', 'loss', 'alpha', 'random_state', 'validation_fraction']
        params_basic = {
            'learning_rate': [0.001, 0.2],
            'n_estimators': [10, 300],
            'max_depth': [1,10],
            'min_samples_split': [10,50],
            'min_samples_leaf': [10,100],
            'max_features': ['sqrt'],
            'subsample': 0.8,
            'loss': ['ls', 'lad', 'huber', 'quantile'],
            'alpha': [0.1, 0.9],
            'random_state': 42,
            'validation_fraction': 0.2
            }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.2,random_state=42)
        def objective(trial):
            param = {
                'learning_rate': trial.suggest_float('learning_rate', params['learning_rate'][0], params['learning_rate'][0]),
                'n_estimators': trial.suggest_int('n_estimators', params['n_estimators'][0], params['n_estimators'][1]),
                'max_depth': trial.suggest_int('max_depth', params['max_depth'][0], params['max_depth'][1]),
                'min_samples_split': trial.suggest_int('min_samples_split', params['min_samples_split'][0], params['min_samples_split'][1]),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', params['min_samples_leaf'][0], params['min_samples_leaf'][1]),
                'max_features': trial.suggest_categorical("max_features", params['max_features']),
                'subsample': trial.suggest_float("subsample", params['subsample'], params['subsample']),
                'loss': trial.suggest_categorical("loss", params['loss']),
                'alpha': trial.suggest_float('alpha', params['alpha'][0], params['alpha'][1]),
                'random_state': trial.suggest_int("random_state", params['random_state'], params['random_state']),
                'validation_fraction': trial.suggest_float("validation_fraction", params['validation_fraction'], params['validation_fraction'])
            }
            gbm = GradientBoostingRegressor(**param)
            gbm.fit(X_train, y_train)
            y_pred = gbm.predict(X_test)
            mse = mean_squared_error(y_test, y_pred)
            return mse
        
        study = optuna.create_study(direction='minimize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        best_params = study.best_params
        gbm_best = GradientBoostingRegressor(**best_params)
        gbm_best.fit(X, y)
        self.model= gbm_best
        self.parameters = best_params
    
    def score(self, X, y):
        preds = np.round(self.model.predict(X))
        return self.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],))
        r2 = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            r2.append(r2_score(y_test,val_pred))

            print(f"{i} Fold scored: {r2[i]}")

        print(f"Mean r2_score {np.mean(r2)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters

    def save(self, model_path="gradient_boosting.joblib"):
        dump(self.model, model_path)
    
    def load(self, model_path):
        self.model = load(model_path)





from sklearn.model_selection import train_test_split,GridSearchCV,KFold
from ..metrics import Metrics
import optuna.visualization as vis
from sklearn.metrics import log_loss, roc_auc_score
import numpy as np
import xgboost as xgb
import optuna
optuna.logging.disable_default_handler()


class XGB(Metrics):
    def __init__(self):
        self.metric = Metrics()
        self.model = None
        self.parameters = None

    def create(self,X,y,params=None,cv=3):
        if params == None:
            params={
                "n_estimators": 50,
            }
        xgb = xgb.XGBClassifier(**params)
        xgb.fit(X,y)
        self.model = xgb
        self.parameters = params

    def create_grid(self, X,y, params=None, cv=3):
        params_columns = ['n_estimators','learning_rate','max_depth',
                          'min_child_weight', 'subsample','colsample_bytree',
                          'gamma','scale_pos_weight','objective','eval_metric']
        params_basic = {
                'n_estimators': [100, 500],  
                'learning_rate': [0.01, 0.1], 
                'max_depth': [1, 7],  
                'min_child_weight': [1, 10],  
                'subsample': [0.8],  
                'colsample_bytree': [0.6, 1.0],  
                'gamma': [0, 0.3],  
                'scale_pos_weight': [1,10], 
                'objective': ['binary:logistic'], 
                'eval_metric': ['logloss', 'auc'],
                'random_state': [42]}
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        mxgb = xgb.XGBClassifier()
        grid_search = GridSearchCV(mxgb,params,cv=cv)
        grid_search.fit(X,y)
        model = grid_search.best_estimator_
        self.model = model
        self.parameters = grid_search.best_params_

    def create_optuna(self,X,y,params=None,n_trials=3, show_plot=False, show_features=False):
        params_columns = ['n_estimators','learning_rate','max_depth', 'random_state',
                          'min_child_weight', 'subsample','colsample_bytree', 'tree_method',
                          'gamma','scale_pos_weight','objective','eval_metric']
        params_basic = {
            'n_estimators': [100, 500],  
            'learning_rate': [0.01, 0.1], 
            'max_depth': [1, 7],  
            'min_child_weight': [1, 10],  
            'subsample': 0.8,  
            'colsample_bytree': [0.6, 1.0],  
            'gamma': [0, 0.3],  
            'scale_pos_weight': [1,10], 
            'objective': ['binary:logistic'], 
            'eval_metric': ['logloss'],
            'tree_method': ['approx'],
            "random_state": 42}
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.2,random_state=42)
        def objective(trial):
            param = {
                'n_estimators': trial.suggest_int('n_estimators', params['n_estimators'][0], params['n_estimators'][1]),
                'learning_rate': trial.suggest_float('learning_rate', params['learning_rate'][0], params['learning_rate'][1]),
                'max_depth': trial.suggest_int('max_depth', params['max_depth'][0], params['max_depth'][1]),
                'min_child_weight': trial.suggest_int('min_child_weight', params['min_child_weight'][0], params['min_child_weight'][1]),
                'subsample': trial.suggest_float("subsample", params['subsample'], params['subsample']),
                'colsample_bytree': trial.suggest_float('colsample_bytree', params['colsample_bytree'][0], params['colsample_bytree'][1]),
                'gamma': trial.suggest_float('gamma', params['gamma'][0], params['gamma'][1]),
                'scale_pos_weight': trial.suggest_float('scale_pos_weight', params['scale_pos_weight'][0], params['scale_pos_weight'][1]),
                'objective': trial.suggest_categorical("objective", params['objective']),
                'tree_method': trial.suggest_categorical("tree_method", params['tree_method']),
                'random_state': trial.suggest_int("random_state", params["random_state"], params["random_state"]),
                'eval_metric': trial.suggest_categorical("eval_metric", params['eval_metric'])
            }
            mxgb = xgb.XGBClassifier(**param)
            mxgb.fit(X_train, y_train)
            preds = mxgb.predict(X_test)
            logloss = log_loss(y_test, preds)
            return logloss
        study = optuna.create_study(direction='minimize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective,n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        best_params = study.best_params
        xgb_best = xgb.XGBClassifier(**best_params)
        xgb_best.fit(X, y)
        self.model = xgb_best
        self.parameters = best_params

    def score(self, X, y):
        preds = np.round(self.xgb.predict(X))
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        return self.xgb.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None, classes=1):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],classes))
        roc = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += np.expand_dims(self.predict(df_test)/n_splits, axis=-1)

            val_pred = self.predict(X_test)
            if classes > 1:
              roc.append(roc_auc_score(y_test,val_pred,multi_class='ovr'))
            else:
              roc.append(roc_auc_score(y_test,val_pred))
            print(f"{i} Fold scored: {roc[i]}")

        print(f"Mean roc score {np.mean(roc)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path='xgb.bin'):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model.load_model(model_path)
    


from sklearn.model_selection import GridSearchCV, train_test_split,KFold
from sklearn.metrics import mean_squared_error,r2_score
from ..metrics_regression import Metrics
import optuna.visualization as vis
import xgboost as xgb
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class XGB(Metrics):
    def __init__(self):
        self.model = None
        self.parameters = None

    def put(self, model):
        self.model = model

    def create(self, X, y, params=None):
        if params == None:
            gbm = xgb.XGBRegressor()
            gbm.fit(X,y)
            self.model = gbm
        else:
            gbm = xgb.XGBRegressor(**params)
            gbm.fit(X,y)
            self.model = gbm
            self.parameters = params

    def create_grid(self, X,y, params=None, cv=3):
        params_columns = ['eval_metric', 'boosting_type', 'objective', 'max_bin',
                          'n_estimators', 'max_depth', 'learning_rate', 'subsample',
                          'colsample_bytree', 'gamma', 'min_child_weight', 'pos_scale_weight',
                          'max_delta_step', 'seed', 'num_boost_round', 'verbosity']
        params_basic = {
            'eval_metric': ['rmse'],
            'boosting_type': ['gblinear'],
            'objective': ['regression'],
            'learning_rate': [1e-4, 1e-1],
            'gamma': [0, 10],
            'min_child_weight': [3, 20],
            'pos_scale_weight': [0.1, 2],
            'max_delta_step': [1, 10],
            'num_leaves': [20,200],
            'n_estimators': [30, 100],
            'max_depth': [5, 32],
            'min_child_samples': [30,70],
            'subsample': [0.3, 0.9],
            'colsample_bytree': [0.5, 1.0],
            'seed': [42],
            'max_bin': [255],
            'num_boost_round': [300],
            'verbosity': -1
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        gbm = xgb.XGBRegressor()
        grid_search = GridSearchCV(gbm, params, cv=cv, n_jobs=-1)
        grid_search.fit(X, y)
        best_model = grid_search.best_estimator_
        self.model = best_model
        self.parameters = grid_search.best_params_

    def create_optuna(self, X, y, params=None, n_trials=5,available_memory_gb=3, show_plot=False, show_features=False):
        rows = len(X)
        memory_row = X.memory_usage(deep=True).sum() / rows
        max_bin = int(np.floor(available_memory_gb * 1024**3 / (rows * memory_row)))

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)
        dtrain = xgb.DMatrix(X_train, label=y_train)
        dtest = xgb.DMatrix(X_test, label=y_test)

        params_columns = ['eval_metric', 'boosting_type', 'objective', 'max_bin', 'tree_method',
                          'n_estimators', 'max_depth', 'learning_rate', 'subsample',
                          'colsample_bytree', 'gamma', 'min_child_weight', 'pos_scale_weight',
                          'max_delta_step', 'seed', 'num_boost_round', 'verbosity']
        params_basic = {
            'eval_metric': ['rmse'],
            'boosting_type': ['gblinear'],
            'objective': ['reg:squarederror'],
            'learning_rate': [1e-4, 1e-1],
            'gamma': [0, 10],
            'min_child_weight': [3, 20],
            'pos_scale_weight': [0.1, 2],
            'max_delta_step': [1, 10],
            'num_leaves': [20,200],
            'n_estimators': [30, 100],
            'max_depth': [5, 32],
            'min_child_samples': [30,70],
            'subsample': [0.3, 0.9],
            'colsample_bytree': [0.5, 1.0],
            'tree_method': ['approx'],
            'seed': 42,
            'max_bin': max_bin,
            'num_boost_round': 300, 
            'verbosity': 0
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]
        
        def objective(trial):
            param = {
                'eval_metric': trial.suggest_categorical("eval_metric", params['eval_metric']),
                'boosting_type': trial.suggest_categorical("boosting_type", params['boosting_type']),
                'objective': trial.suggest_categorical("objective", params['objective']),
                'max_bin': trial.suggest_int("max_bin", params['max_bin'], params['max_bin']),
                'n_estimators': trial.suggest_int('n_estimators', params["n_estimators"][0], params["n_estimators"][1]),
                'max_depth': trial.suggest_int('max_depth', params["max_depth"][0], params["max_depth"][1]),
                'learning_rate': trial.suggest_float('learning_rate', params['learning_rate'][0], params['learning_rate'][1]),
                'subsample': trial.suggest_float('subsample', params['subsample'][0], params['subsample'][1]),
                'colsample_bytree': trial.suggest_float('colsample_bytree', params['colsample_bytree'][0], params['colsample_bytree'][1]),
                'gamma': trial.suggest_float('gamma', params['gamma'][0], params['gamma'][1]),
                'min_child_weight': trial.suggest_int('min_child_weight', params['min_child_weight'][0], params['min_child_weight'][1]),
                'pos_scale_weight': trial.suggest_float('pos_scale_weight', params['pos_scale_weight'][0], params['pos_scale_weight'][1]),
                'max_delta_step': trial.suggest_int('max_delta_step', params['max_delta_step'][0], params['max_delta_step'][1]),
                'seed': trial.suggest_int("seed", params['seed'], params['seed']),
                'tree_method': trial.suggest_categorical("tree_method", params['tree_method']),
                'num_boost_round': trial.suggest_int("num_boost_round", params["num_boost_round"], params["num_boost_round"]),
                'verbosity': trial.suggest_int("verbosity", params['verbosity'], params['verbosity'])
            }
            xgb_model= xgb.train(param, dtrain)
            y_pred = xgb_model.predict(dtest)
            rmse = np.sqrt(mean_squared_error(y_test, y_pred))
            return rmse
    
        study = optuna.create_study(direction='minimize',pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        best_params = study.best_trial.params
        model = xgb.XGBRegressor(**best_params)
        self.model = model.fit(X, y)
        self.parameters = best_params

    def score(self, X, y):
        preds = np.round(self.model.predict(X))
        return self.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],))
        r2 = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            r2.append(r2_score(y_test,val_pred))

            print(f"{i} Fold scored: {r2[i]}")

        print(f"Mean r2_score {np.mean(r2)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path='xgb.bin'):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model.load_model(model_path)


from .cat_automated import CatAutomated
from ..algorithms.complexity_level0 import Complexity_Level0
from ..algorithms.complexity_level1 import Complexity_Level1
from .extra_trees import ExtraTreesAutomated
from .gradient_boosting_automated import GradientBoostingAutomated
from .hist_gradient_automated import HistAutomated
from .lgb_automated import LGBAutomated
from .logistic_regression_automated import LogisticRegressionAutomated
from .random_forest_automated import RandomForestAutomated
from .xgb_automated import XGBAutomated


from sklearn.model_selection import train_test_split, GridSearchCV, KFold
import optuna.integration.lightgbm as lgb
from ..metrics import Metrics
import optuna.visualization as vis
from sklearn.metrics import accuracy_score, roc_auc_score
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class LightLGB(Metrics):
    def __init__(self):
        self.metric = Metrics()
        self.model = None
        self.parameters = None

    def create(self, X, y, params=None):
        if params == None:
            cat = lgb(objective="binary")
            cat.fit(X,y)
            self.model = cat
        else:
            cat = lgb(**params)
            cat.fit(X,y)
            self.model = cat
            self.parameters = params

    def create_grid(self,X,y,params=None,cv=3,available_memory_gb=5):
        rows = len(X)
        memory_row = X.memory_usage(deep=True).sum() / rows
        max_bin = int(np.floor(available_memory_gb * 1024**3 / (rows * memory_row)))

        params_columns = ["max_bin","objective","metric","boosting_type",
                          "lambda_l1","lambda_l2","num_leaves","feature_fraction",
                          "bagging_fraction","bagging_freq","min_child_samples",
                          "max_depth","min_gain_to_split","seed","num_boost_round",
                          "verbosity"]
        params_basic ={
            "max_bin": [max_bin],
            "objective": ["binary"],
            "metric": ["binary_error"],
            "boosting_type": ['gbdt', 'dart'],
            "lambda_l1": [1e-4, 1],
            "lambda_l2": [1e-4,1],
            "num_leaves": [3, 100],
            "feature_fraction": [0.2, 0.8],
            "bagging_freq": [0.1, 0.9],
            "min_child_samples": [2,100],
            "max_depth": [1, 7],
            "min_gain_to_split": [2, 30],
            "seed": [42],
            "num_boost_round": [100],
            "verbosity": [False]
        }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        lgb = lgb()
        grid_search = GridSearchCV(lgb, params, cv=cv)
        grid_search.fit(X, y)
        model = grid_search.best_estimator_
        self.model = model
        self.parameters = grid_search.best_params_

    def create_optuna(self, X, y, params=None, n_trials=10,available_memory_gb=5, show_plot=False, show_features=False):
        ## Calculate processor usage 
        rows = len(X)
        memory_row = X.memory_usage(deep=True).sum() / rows
        max_bin = int(np.floor(available_memory_gb * 1024**3 / (rows * memory_row)))

        params_columns = ["max_bin","objective","metric","boosting_type",
                          "lambda_l1","lambda_l2","num_leaves","feature_fraction",
                          "bagging_freq","min_child_samples","num_boost_round",
                          "max_depth","min_gain_to_split","seed",
                          "verbosity"]
        params_basic ={
            "max_bin": max_bin,
            "objective": ["binary"],
            "metric": "binary_error",
            "boosting_type": ['gbdt', 'dart'],
            "lambda_l1": [0, 1],
            "lambda_l2": [0,1],
            "num_leaves": [3, 100],
            "feature_fraction": [0.2, 0.8],
            "bagging_freq": [0.1, 0.9],
            "min_child_samples": [2,100],
            "max_depth": [1, 7],
            "min_gain_to_split": [2, 30],
            "seed": 42,
            "num_boost_round": 100,
            "verbosity": False
        }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.15, random_state=42)
        dtrain = lgb.Dataset(X_train, label=y_train)
        dval = lgb.Dataset(X_test, label=y_test)
        def objective(trial):
            param = {
                'objective': trial.suggest_categorical('objective', params["objective"]),
                'metric': trial.suggest_categorical('metric', params["metric"]),
                'max_bin': trial.suggest_int("max_bin", params["max_bin"], params["max_bin"]),
                'boosting_type': trial.suggest_categorical("boosting_type", params["boosting_type"]),
                'lambda_l1': trial.suggest_float('lambda_l1', params["lambda_l1"][0], params["lambda_1"][1]),
                'lambda_l2': trial.suggest_float('lambda_l2', params["lambda_l2"][0], params["lambda_2"][1]),
                'num_leaves': trial.suggest_int('num_leaves', params["num_leaves"][0], params["num_leaves"][1]),
                'feature_fraction': trial.suggest_uniform('feature_fraction', params["feature_fraction"][0], params["feature_fraction"][1]),
                'bagging_freq': trial.suggest_int('bagging_freq', params["bagging_freq"][0], params["bagging_freq"][1]),
                'min_child_samples': trial.suggest_int('min_child_samples', params["min_child_samples"][0], params["min_child_samples"][0]),
                'max_depth': trial.suggest_int('max_depth', params["max_depth"][0], params["max_depth"][1]),
                'min_gain_to_split': trial.suggest_loguniform('min_gain_to_split', params["min_gain_to_split"][0], params["min_gain_to_split"][1]),
                'seed': trial.suggest_int("seed", params["seed"], params["seed"]),
                'num_boost_round': trial.suggest_int("num_boost_round", params["num_boost_round"], params["num_boost_round"]),
                'verbosity': params["verbosity"]
            }
            gbm = lgb.train(param, dtrain,valid_sets=dval)
            preds = gbm.predict(X_train)
            pred_labels = np.rint(preds)
            accuracy = accuracy_score(y_train, pred_labels)
            return accuracy

        study = optuna.create_study(direction='maximize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()
            
        dtrain = lgb.Dataset(X_train, label=y_train)
        dval = lgb.Dataset(X_test, label=y_test)
        best_params = study.best_trial.params
        booster = lgb.train(best_params, dtrain, valid_sets=dval,
                            verbose_eval=0)
        self.model=booster
        self.parameters = best_params

    def score(self,X,y):
        preds = self.model.predict(X)
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None, classes=1):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],classes))
        roc = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += np.expand_dims(self.predict(df_test)/n_splits, axis=-1)

            val_pred = self.predict(X_test)
            if classes > 1:
              roc.append(roc_auc_score(y_test,val_pred,multi_class='ovr'))
            else:
              roc.append(roc_auc_score(y_test,val_pred))
            print(f"{i} Fold scored: {roc[i]}")

        print(f"Mean roc score {np.mean(roc)}")
        return predictions
    
    def save(self, model_path='lgb.txt'):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model =lgb.Booster(model_file=model_path)


from sklearn.model_selection import GridSearchCV, train_test_split,KFold
from sklearn.metrics import mean_squared_error,r2_score
from ..metrics_regression import Metrics
import optuna.visualization as vis
from catboost import CatBoostRegressor
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class Cat(Metrics):
    def __init__(self):
        self.model = None
        self.parameters = None

    def put(self, model):
        self.model = model

    def create(self, X, y, params=None):
        if params == None:
            gbm = CatBoostRegressor()
            gbm.fit(X,y)
            self.model = gbm
        else:
            gbm = CatBoostRegressor(**params)
            gbm.fit(X,y)
            self.model = gbm
            self.parameters = params

    def create_grid(self, X,y,params=None, cv=3):
        params_columns = ['depth', 'l2_leaf_reg','border_count','feature_border_type'
                          'bagging_temperature','random_strength','fold_permutation_type'
                          'grow_policy','leaf_estimation_method','random_seed', 'verbose']
        params_basic = {
            'depth': [3,13],
            'l2_leaf_reg': [1, 3],
            'border_count': [32,128],
            'bagging_temperature': [0, 0.7],
            'random_strength': [0, 0.7],
            'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
            'leaf_estimation_method': ['Newton', 'Gradient', 'Exact'],
            "feature_border_type": ["Uniform","MinEntropy"],
            "fold_permutation_block": [3],
            'random_seed': [42],
            'verbose': [0]
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]
        
        regressor = CatBoostRegressor(loss_function='MAE', eval_metric='MAE')
        grid_search = GridSearchCV(regressor, params, cv=cv, scoring='neg_mean_absolute_error', n_jobs=-1)
        grid_search.fit(X,y)
        best_params = grid_search.best_params_
        best_model = grid_search.best_estimator_
        self.model = best_model
        self.parameters = best_params

    def create_optuna(self, X, y, params=None, n_trials=5, show_plot=False, show_features=False):
        params_columns = ['depth', 'l2_leaf_reg','border_count','feature_border_type'
                          'bagging_temperature','random_strength','fold_permutation_type'
                          'grow_policy','leaf_estimation_method','random_seed','verbose']
        params_basic = {
            'depth': [1,5],
            'l2_leaf_reg': [1, 3],
            'border_count': [4,128],
            'bagging_temperature': [0, 0.7],
            'random_strength': [0, 0.7],
            'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
            'leaf_estimation_method': ['Newton', 'Gradient', 'Exact'],
            "feature_border_type": ["Uniform","MinEntropy"],
            "fold_permutation_block": 3,
            'random_seed': 42,
            'verbose': 0
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.2,random_state=42)
        def objective(trial):
            param = {
                  'depth': trial.suggest_int('depth', params['depth'][0], params['depth'][1]),
                  'l2_leaf_reg': trial.suggest_int('l2_leaf_reg', params['l2_leaf_reg'][0], params['l2_leaf_reg'][1]),
                  'border_count': trial.suggest_int('border_count', params['border_count'][0], params['border_count'][1]),
                  'bagging_temperature': trial.suggest_float('bagging_temperature', params['bagging_temperature'][0],params['bagging_temperature'][1]),
                  'random_strength': trial.suggest_float('random_strength', params['random_strength'][0], params['random_strength'][1]),
                  'grow_policy': trial.suggest_categorical("grow_policy", params['grow_policy']),
                  'leaf_estimation_method': trial.suggest_categorical("leaf_estimation_method", params['leaf_estimation_method']),
                  'feature_border_type': trial.suggest_categorical("feature_border_type", params['feature_border_type']),
                  'fold_permutation_block': trial.suggest_int("fold_permutation_block", params['fold_permutation_block'], params['fold_permutation_block']),
                  'random_seed': trial.suggest_int("random_state", params['random_seed'], params['random_seed']),
                  'verbose': trial.suggest_int("verbose", params['verbose'], params['verbose'])
            }
            gbm = CatBoostRegressor(**param)
            gbm.fit(X_train, y_train)
            y_pred = gbm.predict(X_test)
            mse = mean_squared_error(y_test, y_pred)
            return mse
        
        study = optuna.create_study(direction='minimize',pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        best_params = study.best_trial.params
        model = CatBoostRegressor(**best_params)
        self.model = model.fit(X, y)
        self.parameters = best_params
    
    def score(self, X, y):
        preds = np.round(self.model.predict(X))
        return self.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],))
        r2 = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            r2.append(r2_score(y_test,val_pred))

            print(f"{i} Fold scored: {r2[i]}")

        print(f"Mean r2_score {np.mean(r2)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path="cat.cbm"):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model.load_model(model_path)